/*******************************************************************************
 *  Copyright (c) 2010 Association for Decentralized Information Management in
 *  Industry THTH ry.
 *  All rights reserved. This program and the accompanying materials
 *  are made available under the terms of the Eclipse Public License v1.0
 *  which accompanies this distribution, and is available at
 *  http://www.eclipse.org/legal/epl-v10.html
 *
 *  Contributors:
 *      VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard;

import java.io.IOException;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;

import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.adapter.Adapter;
import org.simantics.databoard.adapter.AdapterConstructionException;
import org.simantics.databoard.adapter.AdapterFactory;
import org.simantics.databoard.adapter.RuntimeAdaptException;
import org.simantics.databoard.adapter.RuntimeAdapterConstructionException;
import org.simantics.databoard.annotations.ArgumentImpl;
import org.simantics.databoard.annotations.Arguments;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.BooleanBinding;
import org.simantics.databoard.binding.ByteBinding;
import org.simantics.databoard.binding.DoubleBinding;
import org.simantics.databoard.binding.FloatBinding;
import org.simantics.databoard.binding.IntegerBinding;
import org.simantics.databoard.binding.LongBinding;
import org.simantics.databoard.binding.StringBinding;
import org.simantics.databoard.binding.VariantBinding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.error.RuntimeBindingConstructionException;
import org.simantics.databoard.binding.factory.BindingFactory;
import org.simantics.databoard.binding.factory.DefaultBindingFactory;
import org.simantics.databoard.binding.factory.MutableBindingFactory;
import org.simantics.databoard.binding.impl.BooleanArrayBinding;
import org.simantics.databoard.binding.impl.BooleanBindingDefault;
import org.simantics.databoard.binding.impl.ByteArrayBinding;
import org.simantics.databoard.binding.impl.ByteBindingDefault;
import org.simantics.databoard.binding.impl.DoubleArrayBinding;
import org.simantics.databoard.binding.impl.DoubleBindingDefault;
import org.simantics.databoard.binding.impl.FloatArrayBinding;
import org.simantics.databoard.binding.impl.FloatBindingDefault;
import org.simantics.databoard.binding.impl.IntArrayBinding;
import org.simantics.databoard.binding.impl.IntegerBindingDefault;
import org.simantics.databoard.binding.impl.LongArrayBinding;
import org.simantics.databoard.binding.impl.LongBindingDefault;
import org.simantics.databoard.binding.impl.ObjectVariantBinding;
import org.simantics.databoard.binding.impl.StringArrayBinding;
import org.simantics.databoard.binding.impl.StringBindingDefault;
import org.simantics.databoard.binding.impl.StringVariantBinding;
import org.simantics.databoard.binding.impl.UnsignedByteBinding;
import org.simantics.databoard.binding.impl.UnsignedIntegerBinding;
import org.simantics.databoard.binding.impl.UnsignedLongBinding;
import org.simantics.databoard.binding.mutable.ImmutableVariantBinding;
import org.simantics.databoard.binding.mutable.MutableBooleanBinding;
import org.simantics.databoard.binding.mutable.MutableByteBinding;
import org.simantics.databoard.binding.mutable.MutableDoubleBinding;
import org.simantics.databoard.binding.mutable.MutableFloatBinding;
import org.simantics.databoard.binding.mutable.MutableIntegerBinding;
import org.simantics.databoard.binding.mutable.MutableLongBinding;
import org.simantics.databoard.binding.mutable.MutableStringBinding;
import org.simantics.databoard.binding.mutable.MutableVariantBinding;
import org.simantics.databoard.binding.reflection.BindingRequest;
import org.simantics.databoard.binding.reflection.ReflectionBindingFactory;
import org.simantics.databoard.serialization.DefaultSerializerFactory;
import org.simantics.databoard.serialization.RuntimeSerializerConstructionException;
import org.simantics.databoard.serialization.Serializer;
import org.simantics.databoard.serialization.SerializerConstructionException;
import org.simantics.databoard.serialization.SerializerFactory;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.Datatype;
import org.simantics.databoard.util.DataValueUtil;

/**
 * This ia a facade class for the binding services.
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class Bindings {

	// Repositories
    public static final Map<Datatype, Binding> mutableBindingRepository = Collections.synchronizedMap( new HashMap<Datatype, Binding>() );
    public static final Map<Datatype, Binding> defaultBindingRepository = Collections.synchronizedMap( new HashMap<Datatype, Binding>() );
    public static final Map<BindingRequest, Binding> reflectionBindingRepository = Collections.synchronizedMap( new HashMap<BindingRequest, Binding>() );
    public static final Map<Binding, Serializer> serializerRepository = Collections.synchronizedMap( new HashMap<Binding, Serializer>() ); 
	
    // Factories
	public static final BindingFactory mutableBindingFactory = new MutableBindingFactory( mutableBindingRepository );
	public static final BindingFactory defaultBindingFactory = new DefaultBindingFactory( defaultBindingRepository );
    public static final ReflectionBindingFactory reflectionBindingFactory = new ReflectionBindingFactory( reflectionBindingRepository ); 
    public static final SerializerFactory defaultSerializationFactory = new DefaultSerializerFactory( serializerRepository ); 
	public static final AdapterFactory adapterFactory = new AdapterFactory();

	// Bindings to java primitives, String, Integer, Boolean, Byte, Long, Double, Float, Object, boolean[], byte[], int[], long[], float[], double[]
    // These bindings have no value range or unit type 
	public static final StringBinding  STRING         = new StringBindingDefault( Datatypes.STRING );
	public static final IntegerBinding INTEGER        = new IntegerBindingDefault( Datatypes.INTEGER );	
	public static final BooleanBinding BOOLEAN        = new BooleanBindingDefault( Datatypes.BOOLEAN ); 
	public static final ByteBinding    BYTE           = new ByteBindingDefault( Datatypes.BYTE ); 
	public static final LongBinding    LONG           = new LongBindingDefault( Datatypes.LONG ); 
	public static final DoubleBinding  DOUBLE         = new DoubleBindingDefault( Datatypes.DOUBLE ); 
	public static final FloatBinding   FLOAT          = new FloatBindingDefault( Datatypes.FLOAT ); 
	public static final ObjectVariantBinding OBJECT   = new ObjectVariantBinding(Bindings.reflectionBindingRepository); 
	public static final VariantBinding VARIANT        = new ImmutableVariantBinding();
	public static final VariantBinding STR_VARIANT    = new StringVariantBinding();
	public static final ArrayBinding   BOOLEAN_ARRAY  = BooleanArrayBinding.createFrom(new ArrayType(Datatypes.BOOLEAN)); 
	public static final ArrayBinding   BYTE_ARRAY     = ByteArrayBinding.createFrom(new ArrayType(Datatypes.BYTE)); 
	public static final ArrayBinding   INT_ARRAY      = IntArrayBinding.createFrom(new ArrayType(Datatypes.INTEGER)); 
	public static final ArrayBinding   LONG_ARRAY     = LongArrayBinding.createFrom(new ArrayType(Datatypes.LONG)); 
	public static final ArrayBinding   FLOAT_ARRAY    = FloatArrayBinding.createFrom(new ArrayType(Datatypes.FLOAT)); 
	public static final ArrayBinding   DOUBLE_ARRAY   = DoubleArrayBinding.createFrom(new ArrayType(Datatypes.DOUBLE));
	public static final ArrayBinding   STRING_ARRAY   = StringArrayBinding.createFrom(new ArrayType(Datatypes.STRING));	

	public static final IntegerBinding UNSIGNED_INTEGER         = new UnsignedIntegerBinding.Immutable( Datatypes.INTEGER );
	public static final ByteBinding    UNSIGNED_BYTE            = new UnsignedByteBinding.Immutable( Datatypes.BYTE );
	public static final LongBinding    UNSIGNED_LONG            = new UnsignedLongBinding.Immutable( Datatypes.LONG );
	
	public static final StringBinding  MUTABLE_STRING           = new MutableStringBinding( Datatypes.STRING );
	public static final IntegerBinding MUTABLE_INTEGER          = new MutableIntegerBinding( Datatypes.INTEGER );	
	public static final BooleanBinding MUTABLE_BOOLEAN          = new MutableBooleanBinding( Datatypes.BOOLEAN ); 
	public static final ByteBinding    MUTABLE_BYTE             = new MutableByteBinding( Datatypes.BYTE ); 
	public static final LongBinding    MUTABLE_LONG             = new MutableLongBinding( Datatypes.LONG ); 
	public static final FloatBinding   MUTABLE_FLOAT            = new MutableFloatBinding( Datatypes.FLOAT ); 
	public static final DoubleBinding  MUTABLE_DOUBLE           = new MutableDoubleBinding( Datatypes.DOUBLE ); 
	public static final VariantBinding MUTABLE_VARIANT          = new MutableVariantBinding();
	
	public static final IntegerBinding MUTABLE_UNSIGNED_INTEGER = new UnsignedIntegerBinding.Mutable( Datatypes.INTEGER );
	public static final ByteBinding    MUTABLE_UNSIGNED_BYTE    = new UnsignedByteBinding.Mutable( Datatypes.BYTE );
	public static final LongBinding    MUTABLE_UNSIGNED_LONG    = new UnsignedLongBinding.Mutable( Datatypes.LONG );	
	
    /**
     * Get a binding to a DataType that is completely mutable java class. 
     * 
	 * DataType           | Class of the bound instance
	 * ===================|==================
	 * UnionType          | GenericBinding.TaggedObject.class
	 * OptionType         | ValueContainer.class
	 * RecordType         | Object[].class
	 * BooleanType        | MutableBoolean.class
	 * DoubleType         | MutableDouble.class
	 * FloatType          | MutableFloat.class
	 * ByteType           | MutableByte.class
	 * IntegerType        | MutableInt.class
	 * LongType           | MutableLong.class
	 * StringType         | ValueContainer.class
	 * ArrayType          | ArrayList.class
	 * MapType            | TreeMap.class
	 * VariantType        | MutableVariant.class
	 * 
	 * Note, requesting a binding with this convenience method stores the 
	 * binding and the type with strong reference, thus preventing garbage 
	 * collection. To allow garbage collection, please use another instance of 
	 * GenericBindingFactory and binding repository (Map<Datatype, Binding>). <p>
     *
	 * @param type
	 * @return binding binding to a mutable class 
     */    
    @SuppressWarnings("unchecked")
	public static <T extends Binding> T getMutableBinding(Datatype type) {
    	try {
    		Binding binding = mutableBindingRepository.get(type);
    		if (binding!=null) return (T) binding;
        	synchronized(mutableBindingRepository) {
    			return (T) mutableBindingFactory.getBinding(type);
        	}
		} catch (BindingConstructionException e) {
			// Unexpected - if error is thrown there is fault in GenericBindingScheme
			throw new RuntimeBindingConstructionException(e);
		}
    }
    
    /**
     * Get a binding that uses default java classes, such as Integer.class, or
     * byte[].class. The result is almost always immutable. This method is more
     * efficient than generic bindings (above).
     * 
     * BooleanType        | Boolean.class
     * ByteType           | Byte.class
     * FloatType          | Float.class
     * DoubleType         | Double.class
     * IntegerType        | Int.class
     * LongType           | Long.class
     * StringType         | String.class
     * UnionType          | TaggedObject.class
     * OptionType         | ValueContainer.class
     * RecordType         | Object[].class
     * MapType            | TreeMap.class
     * VariantType        | Variant.class
     * ArrayType(Boolean) | boolean[].class
     * ArrayType(Byte)    | byte[].class
     * ArrayType(Integer) | int[].class
     * ArrayType(Long)    | long[].class
     * ArrayType(Float)   | float[].class
     * ArrayType(Double)  | double[].class
     * ArrayType(Byte)    | byte[].class
     * ArrayType( T )     | Object[].class
     * 
	 * Note, requesting a binding with this convenience method stores the 
	 * binding and the type with strong reference, thus preventing garbage 
	 * collection. To allow garbage collection, please use another instance of 
	 * DefaultBindingFactory and binding repository (Map<Datatype, Binding>). <p>
     *
	 * @param type
	 * @return binding binding to a mutable class 
     */    
    @SuppressWarnings("unchecked")
	public static <T extends Binding> T getBinding(Datatype type) {
    	try {
    		Binding binding = defaultBindingRepository.get(type);
    		if (binding!=null) return (T) binding;
        	synchronized(defaultBindingRepository) {
    			return (T) defaultBindingFactory.getBinding(type);
        	}
        } catch (BindingConstructionException e) {
			// Unexpected - if error is thrown there is fault in DefaultBindingScheme
			throw new RuntimeBindingConstructionException(e);
		}
    }    
	
	/**
	 * Get a binding to a Java Class. Type details can be modified with annotations.
	 * Please see the package org.simantics.databoard.annotations. 
	 * <p>
	 *  
	 * The whether the result binding is a completely mutable or not depends on the
	 * provided classes. For instance, fields such as Boolean, Integer, Long
	 * are not mutable, instead MutableBoolean, MutableInteger and MutableLong are.
	 * The length of Object[] is not mutable, but length of List<Object> is. <p>
	 * 
	 * Note, requesting a binding with this convenience method stores the 
	 * binding and the class with strong reference, thus preventing garbage 
	 * collection. To allow garbage collection, please use another instance of 
	 * BindingFactory and binding repository (Map<BindingRequest, Binding>). <p> 
	 * 
	 * @see ReflectionBindingFactory  
	 * @param clazz
	 * @return binding
	 * @throws BindingConstructionException
	 */
    @SuppressWarnings("unchecked")
	public static <T extends Binding> T getBinding(Class<?> clazz)
    throws BindingConstructionException
    {
    	BindingRequest request = new BindingRequest( clazz );
    	Binding binding = reflectionBindingRepository.get( request );
    	if (binding!=null);
    	synchronized(reflectionBindingFactory) {
    		binding = reflectionBindingFactory.construct(request);
    	}
		return (T) binding;
    }
    
	/**
	 * Get a binding to a Java Class. Use this method to acquire class 
	 * parameters for a generics class. <p>
	 * 
	 * Example 1: 
	 * 
	 *    Binding binding = Bindings.getBinding(Map.class, String.class, Integer.class);
	 *    Map<String, Integer> map = (Map<String, Integer>) binding.createDefault();
	 *    
	 * Example 2:
	 *    
     *  Binding d = Bindings.getBinding(List.class, Integer.class);
	 *  List<Integer> list = (List<Integer>) d.createRandom(5);
	 *    
	 * Example 3:
	 *    
     *  Binding d = Bindings.getBinding(List.class, List.class, Integer.class);
	 *  List<List<Integer>> list = (List<List<Integer>>) d.createRandom(5);
	 * 
	 * @see ReflectionBindingFactory  
	 * @param clazz
	 * @return binding
	 * @throws BindingConstructionException
	 */
    @SuppressWarnings("unchecked")
	public static <T extends Binding> T getBinding(Class<?> clazz, Class<?>...parameters)
    throws BindingConstructionException
    {
    	Arguments args = new ArgumentImpl(parameters);
    	BindingRequest request = new BindingRequest( clazz, args );
    	Binding binding = reflectionBindingRepository.get( request );
    	if (binding!=null);
    	synchronized(reflectionBindingFactory) {
    		binding = reflectionBindingFactory.construct(request);
    	}
		return (T) binding;
    }    
    
    /**
	 * Read binding and type from a class. DataType details and parameters
	 * are read as annotations placed in the class.  
	 * (See org.simantics.databoard.annotations)
	 * <p>
	 * As an exception, in the subclasses of {@link Throwable}, the fields of
	 * Throwable are omited.
	 * <p>
	 * This method is used when the caller is 100% sure that binding will be 
	 * constructed without exceptions. Such classes
	 * are all primitive types (Double, Integer, etc, arrays, DataType, ...)
	 * 
	 * This method is unchecked if binding construction to the clazz cannot be trusted.
	 * If construction fails, a RuntimeException is thrown.
	 * 
	 * @param clazz
	 * @return binding
	 * @throws RuntimeBindingConstructionException
     */
	@SuppressWarnings("unchecked")
	public static <T extends Binding> T getBindingUnchecked(Class<?> clazz)
    throws RuntimeBindingConstructionException
    {
		try {
	    	return (T) getBinding(clazz);
		} catch (BindingConstructionException e) {
			throw new RuntimeBindingConstructionException(e);
		}
    }
	
	/**
	 * Get serializer that follows Databoard serialization spec.
	 *  
	 * @param binding
	 * @return serializer
	 * @throws SerializerConstructionException
	 */
	public static Serializer getSerializer(Binding binding) throws SerializerConstructionException {
		return defaultSerializationFactory.construct(binding);
	}

	/**
	 * Get serializer that follows Databoard serialization spec.
	 * 
	 * @param binding
	 * @return serializer
	 * @throws RuntimeSerializerConstructionException
	 */
	public static Serializer getSerializerUnchecked(Binding binding) throws RuntimeSerializerConstructionException {
		try {
			return defaultSerializationFactory.construct(binding);
		} catch (SerializerConstructionException e) {
			throw new RuntimeSerializerConstructionException(e);
		}
	}

	/**
	 * Get serializer that follows Databoard serialization spec.
	 *  
	 * @param clazz
	 * @return serializer
	 * @throws SerializerConstructionException
	 */
	public static Serializer getSerializer(Class<?> clazz) throws SerializerConstructionException {
		try {
			Binding binding = getBinding(clazz);
			return defaultSerializationFactory.construct(binding);
		} catch (BindingConstructionException e) {
			throw new SerializerConstructionException( e );
		}
	}

	/**
	 * Get serializer that follows Databoard serialization spec.
	 * 
	 * @param clazz
	 * @return serializer serializer
	 * @throws RuntimeSerializerConstructionException
	 */
	public static Serializer getSerializerUnchecked(Class<?> clazz) throws RuntimeSerializerConstructionException {
		try {
			Binding binding = getBinding(clazz);
			return defaultSerializationFactory.construct(binding);
		} catch (SerializerConstructionException e) {
			throw new RuntimeSerializerConstructionException(e);
		} catch (BindingConstructionException e) {
			throw new RuntimeSerializerConstructionException( new SerializerConstructionException(e) );
		}
	}	
    /**
     * Create an adapter that adapts two bindings of the same
     * data type.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getAdapter(Binding domain, Binding range)
    throws AdapterConstructionException
    {
    	return adapterFactory.getAdapter(domain, range, false, false);
    }
    
    /**
     * Create an adapter that adapts between two bindings of the same
     * data type.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getAdapterUnchecked(Binding domain, Binding range)
    throws RuntimeAdapterConstructionException
    {
    	try {
			return adapterFactory.getAdapter(domain, range, false, false);
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdapterConstructionException(e);
		}
    }

    /**
     * Create a type adapter that adapts instances from one DataType to 
     * another. Type Adapter does the following conversions: 
     * 
     *    o Number Types, e.g. long -> double
     *    o Unit Types, e.g. mph -> km/h
     *    o Record Types, for each field of the range, there must be equal in domain 
     *    o Union Types, for each tag type of the domain, there must be equal in range
     * 
     * {@link AdaptException} is thrown at runtime, if number values are 
     * not compatible, e.g. converting value 500 from Long to Byte.
     * Note, there is also a possibility of precision loss, e.g. when 
     * converting from double to int.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getTypeAdapter(Binding domain, Binding range)
    throws AdapterConstructionException
    {
    	return adapterFactory.getAdapter(domain, range, true, false);
    }

    /**
     * Create a type adapter that adapts instances from one DataType to 
     * another. Type Adapter does the following conversions: 
     * 
     *    o Number Types, e.g. long -> double
     *    o Unit Types, e.g. mph -> km/h
     *    o Record Types, for each field of the range, there must be equal in domain 
     *    o Union Types, for each tag type of the domain, there must be equal in range
     * 
     * {@link AdaptException} is thrown at runtime, if number values are 
     * not compatible, e.g. converting value 500 from Long to Byte.
     * Note, there is also a possibility of precision loss, e.g. when 
     * converting from double to int.
     * 
     * @param domain binding of the source instance
     * @param range binding of the result instance
     * @return result adapter
     * @throws AdapterConstructionException 
     */
    public static Adapter getTypeAdapterUnchecked(Binding domain, Binding range)
    {
    	try {
			return adapterFactory.getAdapter(domain, range, true, false);
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdapterConstructionException(e);
		}
    }
    
    /**
     * Adapt a value of one type to another. 
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object adapt(Object value, Binding domain, Binding range)
    throws AdaptException
    {
    	try {
			return adapterFactory.getAdapter(domain, range, true, false).adapt(value);
		} catch (AdapterConstructionException e) {
			throw new AdaptException(e);
		}
    }
    
    /**
     * Adapt a value of one type to another
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object adaptUnchecked(Object value, Binding domain, Binding range)
    throws RuntimeAdapterConstructionException, RuntimeAdaptException
    {
    	try {
    		return  adapterFactory.getAdapter(domain, range, true, false).adaptUnchecked(value);
		} catch (RuntimeAdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e.getCause()));
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e));
		}
    }    
    
    /**
     * Clone a value to another type to another. Bindings that handle immutable values
     * may return the same instance, others will guarantee a complete copy.  
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object clone(Object value, Binding domain, Binding range)
    throws AdaptException
    {
    	try {
			return adapterFactory.getAdapter(domain, range, true, true).adapt(value);
		} catch (AdapterConstructionException e) {
			throw new AdaptException(e);
		}
    }
    

    /**
     * Clone a value of one binding to another. Bindings that handle immutable values
     * may return the same instance, others will guarantee a complete copy.
     * 
     * @param value
     * @param domain
     * @param range
     * @return adapted value
     * @throws AdapterConstructionException
     * @throws AdaptException
     */
    public static Object cloneUnchecked(Object value, Binding domain, Binding range)
    throws RuntimeAdapterConstructionException, RuntimeAdaptException
    {
    	try {
			return adapterFactory.getAdapter(domain, range, true, true).adapt(value);
		} catch (AdaptException e) {
			throw new RuntimeAdaptException(e);
		} catch (RuntimeAdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e.getCause()));
		} catch (AdapterConstructionException e) {
			throw new RuntimeAdaptException(new AdaptException(e));
		}
    }
    
    /**
     * Compares its two data values for order.  Returns a negative integer,
     * zero, or a positive integer as the first argument is less than, equal
     * to, or greater than the second.<p>
     * 
     * DataTypes of b1 and b2 are not equal then data types are compared. <p>
     *
     * The comparison function is defined at 
     * http://dev.simantics.org/index.php/Org.simantics.datatype_Manual#CompareTo_and_Equals <p>
     * 
     * Note, comparing 2 different number types will not result a value comparison.
     * Instead values have the following type precedence ByteType, IntegerType, LongType,
     * FloatType, and the highest DoubleType. <p>
     *
     * @param b1 Binding of o1 
     * @param o1 the first object to be compared.
     * @param b2 Binding of o2
     * @param o2 the second object to be compared.
     * @return a negative integer, zero, or a positive integer as the
     * 	       first argument is less than, equal to, or greater than the
     *	       second.
     * @throws BindingException if object cannot be handled by a binding
     */
    public static int compare(Binding b1, Object o1, Binding b2, Object o2) 
    throws BindingException
    {
    	return DataValueUtil.compare(b1, o1, b2, o2);
    }
    
    /**
     * Compare two data values for equality. <p>
     * 
     * Note, comparing 2 different number types will not result a value comparison.
     * Instead values have the following type precedence ByteType, IntegerType, LongType,
     * FloatType, and the highest DoubleType. <p>
     * 
     * @param b1
     * @param o1
     * @param b2
     * @param o2
     * @return true if equal
     * @throws BindingException
     */
    public static boolean equals(Binding b1, Object o1, Binding b2, Object o2)
    throws BindingException
    {
    	return DataValueUtil.equals(b1, o1, b2, o2);
    }
    
    public static Comparator<Object> createComparator(final Binding b1, final Binding b2)
    {
    	return DataValueUtil.createComparator(b1, b2);
    }
    
    /**
     * Print the content of an object as a structure.
     * Utility function for debug purposes.
     * 
     * @param o
     * @return content
     */
    public static String toString(Object o) {
    	try {
			Binding b = Bindings.getBinding( o.getClass() );
			return b.printValueDefinition(o, true);
		} catch (BindingConstructionException e) {
			return "<error "+e.getClass().getName()+" "+e.getMessage()+">";
		} catch (IOException e) {
			return "<error "+e.getClass().getName()+" "+e.getMessage()+">";
		} catch (BindingException e) {
			return "<error "+e.getClass().getName()+" "+e.getMessage()+">";
		}
    }
	
}

