package org.simantics.databoard.binding.reflection;

import java.lang.annotation.Annotation;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.lang.reflect.TypeVariable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.Datatypes;
import org.simantics.databoard.annotations.ArgumentImpl;
import org.simantics.databoard.annotations.Arguments;
import org.simantics.databoard.annotations.Length;
import org.simantics.databoard.annotations.MIMEType;
import org.simantics.databoard.annotations.Optional;
import org.simantics.databoard.annotations.Pattern;
import org.simantics.databoard.annotations.Range;
import org.simantics.databoard.annotations.Referable;
import org.simantics.databoard.annotations.Union;
import org.simantics.databoard.annotations.Unit;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.OptionalBinding;
import org.simantics.databoard.binding.error.BindingConstructionException;
import org.simantics.databoard.binding.error.RuntimeBindingConstructionException;
import org.simantics.databoard.binding.impl.ArrayListBinding;
import org.simantics.databoard.binding.impl.BooleanArrayBinding;
import org.simantics.databoard.binding.impl.ByteArrayBinding;
import org.simantics.databoard.binding.impl.ByteBindingDefault;
import org.simantics.databoard.binding.impl.DoubleArrayBinding;
import org.simantics.databoard.binding.impl.DoubleBindingDefault;
import org.simantics.databoard.binding.impl.FloatArrayBinding;
import org.simantics.databoard.binding.impl.FloatBindingDefault;
import org.simantics.databoard.binding.impl.HashMapBinding;
import org.simantics.databoard.binding.impl.HashSetBinding;
import org.simantics.databoard.binding.impl.IntArrayBinding;
import org.simantics.databoard.binding.impl.IntegerBindingDefault;
import org.simantics.databoard.binding.impl.LinkedListBinding;
import org.simantics.databoard.binding.impl.LongArrayBinding;
import org.simantics.databoard.binding.impl.LongBindingDefault;
import org.simantics.databoard.binding.impl.ObjectArrayBinding;
import org.simantics.databoard.binding.impl.ObjectVariantBinding;
import org.simantics.databoard.binding.impl.OptionalBindingDefault;
import org.simantics.databoard.binding.impl.StringBindingDefault;
import org.simantics.databoard.binding.impl.TreeMapBinding;
import org.simantics.databoard.binding.impl.TreeSetBinding;
import org.simantics.databoard.binding.impl.UnsignedByteBinding;
import org.simantics.databoard.binding.impl.UnsignedIntegerBinding;
import org.simantics.databoard.binding.impl.UnsignedLongBinding;
import org.simantics.databoard.binding.mutable.MutableByteBinding;
import org.simantics.databoard.binding.mutable.MutableDoubleBinding;
import org.simantics.databoard.binding.mutable.MutableFloatBinding;
import org.simantics.databoard.binding.mutable.MutableIntegerBinding;
import org.simantics.databoard.binding.mutable.MutableLongBinding;
import org.simantics.databoard.binding.mutable.MutableStringBinding;
import org.simantics.databoard.binding.mutable.MutableVariant;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.primitives.MutableBoolean;
import org.simantics.databoard.primitives.MutableByte;
import org.simantics.databoard.primitives.MutableDouble;
import org.simantics.databoard.primitives.MutableFloat;
import org.simantics.databoard.primitives.MutableInteger;
import org.simantics.databoard.primitives.MutableLong;
import org.simantics.databoard.primitives.MutableString;
import org.simantics.databoard.primitives.UnsignedByte;
import org.simantics.databoard.primitives.UnsignedInteger;
import org.simantics.databoard.primitives.UnsignedLong;
import org.simantics.databoard.type.ArrayType;
import org.simantics.databoard.type.BooleanType;
import org.simantics.databoard.type.ByteType;
import org.simantics.databoard.type.Component;
import org.simantics.databoard.type.DoubleType;
import org.simantics.databoard.type.FloatType;
import org.simantics.databoard.type.IntegerType;
import org.simantics.databoard.type.LongType;
import org.simantics.databoard.type.MapType;
import org.simantics.databoard.type.OptionalType;
import org.simantics.databoard.type.RecordType;
import org.simantics.databoard.type.StringType;
import org.simantics.databoard.type.UnionType;
import org.simantics.databoard.util.ArrayUtils;
import org.simantics.databoard.util.RangeException;

/**
 * Type Factory constructs data types from reflection requests.
 * Successfully constructed types are placed in the repository that was given 
 * at construction time.
 *  
 * @author Toni Kalajainen
 */
public class ReflectionBindingFactory {
	
	/**
	 * Map of failed constructions. 
	 */
	Map<BindingRequest, BindingConstructionException> failures = new HashMap<BindingRequest, BindingConstructionException>();
	
	/**
	 * Map that contains in incomplete constructions.
	 */
	Map<BindingRequest, Binding> inprogress = new HashMap<BindingRequest, Binding>();
	
	/**
	 * Repository where successful constructions are placed. 
	 */
	Map<BindingRequest, Binding> repository;	

	/**
	 * Construct a new reflection binding factory 
	 */
	public ReflectionBindingFactory() {
		this.repository = new HashMap<BindingRequest, Binding>(); 
	}
	
	/**
	 * Construct a new reflection binding factory that places constructed 
	 * bindings into user given repository.
	 * 
	 * @param repository
	 */
	public ReflectionBindingFactory(Map<BindingRequest, Binding> repository) {
		this.repository = repository;
	}
	
	public Map<BindingRequest, Binding> getRepository() {
		return repository;
	}
	
	/**
	 * Constructs a binding to comply to class request.
	 * This is the method sub-classes implement. 
	 * The implementation should use the inprogress -map for construction of 
	 * bindings that have component types.
	 * 
	 *  e.g. 
	 *   inprogress.put(request, notCompletelyConstructedBinding);
	 *   Binding componentBinding = construct( componentRequest );
	 *   notCompletelyConstructedBinding.setChild( componentBinding );
	 *   inprogress.remove(request);
	 *   
	 * try-finally is not needed.
	 * 
	 * @param request
	 * @return
	 * @throws BindingConstructionException
	 * @throws RangeException
	 */
	protected Binding doConstruct(BindingRequest request) throws BindingConstructionException, RangeException
	{
		// Optional
    	if(request.hasAnnotation(Optional.class))
    	{    		
	    	Optional optional = request.getAnnotation(Optional.class); 
	    	Annotation[] newAnnotations = ArrayUtils.dropElements(request.annotations, optional);
    		BindingRequest componentRequest = new BindingRequest(request.clazz, newAnnotations);
    		OptionalType type = new OptionalType();
    		OptionalBinding binding = new OptionalBindingDefault(type, null);
			inprogress.put(request, binding);
			binding.componentBinding = construct( componentRequest );
			type.componentType = binding.componentBinding.type();
			inprogress.remove(request);	    		
    		// Its complete, store to repository
			repository.put(request, binding);			
			return binding;
    	}
    	
    	// Primitive
		{  		
			Range range = request.getAnnotation(Range.class);
			Unit unit = request.getAnnotation(Unit.class);
			
			if (request.clazz == Integer.class || request.clazz == int.class || request.clazz == MutableInteger.class || UnsignedInteger.class.isAssignableFrom(request.clazz)) {
				Binding binding = null; 
				if (range==null && unit==null) {
					if (request.clazz == int.class) binding = Bindings.INTEGER;
					if (request.clazz == Integer.class) binding = Bindings.INTEGER;
					if (request.clazz == MutableInteger.class) binding = Bindings.MUTABLE_INTEGER;
					if (request.clazz == UnsignedInteger.Mutable.class) binding = Bindings.MUTABLE_UNSIGNED_INTEGER;
					if (request.clazz == UnsignedInteger.Immutable.class) binding = Bindings.UNSIGNED_INTEGER;
				} else {
					IntegerType type = new IntegerType();
					type.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					type.unit = unit==null ? null : unit.value();

					if (request.clazz == int.class) binding = new IntegerBindingDefault(type);
					if (request.clazz == Integer.class) binding = new IntegerBindingDefault(type);
					if (request.clazz == MutableInteger.class) binding = new MutableIntegerBinding(type);
					if (request.clazz == UnsignedInteger.Mutable.class) binding = new UnsignedIntegerBinding.Mutable(type);
					if (request.clazz == UnsignedInteger.Immutable.class) binding = new UnsignedIntegerBinding.Immutable(type);
				}
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
				
				repository.put(request, binding);
				return binding;
			}
				
			if (request.clazz == Byte.class || request.clazz == byte.class || request.clazz == MutableByte.class || UnsignedByte.class.isAssignableFrom(request.clazz)) {
				Binding binding = null; 
				if (range==null && unit==null) {
					if (request.clazz == byte.class) binding = Bindings.BYTE;
					if (request.clazz == Byte.class) binding = Bindings.BYTE;
					if (request.clazz == MutableByte.class) binding = Bindings.MUTABLE_BYTE;
					if (request.clazz == UnsignedByte.Mutable.class) binding = Bindings.MUTABLE_UNSIGNED_BYTE;
					if (request.clazz == UnsignedByte.Immutable.class) binding = Bindings.UNSIGNED_BYTE;
				} else {
					ByteType type = new ByteType();
					type.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					type.unit = unit==null ? null : unit.value();

					if (request.clazz == byte.class) binding = new ByteBindingDefault(type);
					if (request.clazz == Byte.class) binding = new ByteBindingDefault(type);
					if (request.clazz == MutableByte.class) binding = new MutableByteBinding(type);
					if (request.clazz == UnsignedByte.Mutable.class) binding = new UnsignedByteBinding.Mutable(type);
					if (request.clazz == UnsignedByte.Immutable.class) binding = new UnsignedByteBinding.Immutable(type);
				}
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
				
				repository.put(request, binding);			
				return binding;
			}			
			
			if (request.clazz == Long.class || request.clazz == long.class || request.clazz == MutableLong.class || UnsignedLong.class.isAssignableFrom(request.clazz)) {
				Binding binding = null; 
				if (range==null && unit==null) {
					if (request.clazz == long.class) binding = Bindings.LONG;
					if (request.clazz == Long.class) binding = Bindings.LONG;
					if (request.clazz == MutableLong.class) binding = Bindings.MUTABLE_LONG;
					if (request.clazz == UnsignedLong.Mutable.class) binding = Bindings.MUTABLE_UNSIGNED_LONG;
					if (request.clazz == UnsignedLong.Immutable.class) binding = Bindings.UNSIGNED_LONG;
				} else {
					LongType type = new LongType();
					type.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					type.unit = unit==null ? null : unit.value();

					if (request.clazz == long.class) binding = new LongBindingDefault(type);
					if (request.clazz == Long.class) binding = new LongBindingDefault(type);
					if (request.clazz == MutableLong.class) binding = new MutableLongBinding(type);
					if (request.clazz == UnsignedLong.Mutable.class) binding = new UnsignedLongBinding.Mutable(type);
					if (request.clazz == UnsignedLong.Immutable.class) binding = new UnsignedLongBinding.Immutable(type);
				}
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
				
				repository.put(request, binding);			
				return binding;
			}
			
			if (request.clazz == Float.class || request.clazz == float.class || request.clazz == MutableFloat.class) {
				Binding binding = null; 
				if (range==null && unit==null) {
					if (request.clazz == float.class) binding = Bindings.FLOAT;
					if (request.clazz == Float.class) binding = Bindings.FLOAT;
					if (request.clazz == MutableFloat.class) binding = Bindings.MUTABLE_FLOAT;
				} else {
					FloatType type = new FloatType();
					type.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					type.unit = unit==null ? null : unit.value();

					if (request.clazz == float.class) binding = new FloatBindingDefault(type);
					if (request.clazz == Float.class) binding = new FloatBindingDefault(type);
					if (request.clazz == MutableFloat.class) binding = new MutableFloatBinding(type);
				}
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
				
				repository.put(request, binding);			
				return binding;
			}
			
			if (request.clazz == Double.class || request.clazz == double.class || request.clazz == MutableDouble.class) {
				Binding binding = null; 
				if (range==null && unit==null) {
					if (request.clazz == double.class) binding = Bindings.DOUBLE;
					if (request.clazz == Double.class) binding = Bindings.DOUBLE;
					if (request.clazz == MutableDouble.class) binding = Bindings.MUTABLE_DOUBLE;
				} else {
					DoubleType type = new DoubleType();
					type.setRange( range==null ? null : org.simantics.databoard.util.Range.valueOf(range.value()) );
					type.unit = unit==null ? null : unit.value();

					if (request.clazz == double.class) binding = new DoubleBindingDefault(type);
					if (request.clazz == Double.class) binding = new DoubleBindingDefault(type);
					if (request.clazz == MutableDouble.class) binding = new MutableDoubleBinding(type);
				}
				
				repository.put(request, binding);			
				return binding;
			}

			if (request.clazz == Boolean.class || request.clazz == boolean.class || request.clazz == MutableBoolean.class) {
				BooleanType type = Datatypes.BOOLEAN;
				Binding binding = null;
	    		if (request.clazz == Boolean.class || request.clazz == boolean.class) binding = Bindings.BOOLEAN;
	    		if (request.clazz == MutableBoolean.class) binding = Bindings.MUTABLE_BOOLEAN;    		
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
				
				repository.put(request, binding);			
				return binding;
			}

			if (request.clazz == String.class || request.clazz == MutableString.class) {
				Length length = request.getAnnotation(Length.class);
				MIMEType mimeType = request.getAnnotation(MIMEType.class);
				Pattern pattern = request.getAnnotation(Pattern.class);
				Binding binding = null;
				if (length==null && mimeType==null && pattern==null) {
					if (request.clazz == String.class) binding = Bindings.STRING;
					if (request.clazz == MutableString.class) binding = Bindings.MUTABLE_STRING;
					} else {
						StringType type = new StringType();
						type.setLength( length==null ? null : org.simantics.databoard.util.Range.valueOf( length.value()[0] ) );
						type.mimeType = mimeType==null ? null : mimeType.value();
						type.pattern = pattern==null ? null : pattern.value();					
					if (request.clazz == String.class) binding = new StringBindingDefault(type);
					if (request.clazz == MutableString.class) binding = new MutableStringBinding(type);
					}
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
				
	            repository.put(request, binding);                        
				return binding;
			}
		}
    	
		
        // Array or List type
		{
			if(request.clazz.isArray() || request.clazz.equals(ArrayList.class) || request.clazz.equals(LinkedList.class) || List.class.isAssignableFrom(request.clazz) ) {
        		Length lengthAnnotation = request.getAnnotation(Length.class);
        		Arguments argumentsAnnotation = request.getAnnotation(Arguments.class);
        		Annotation[] componentAnnotations = request.dropAnnotations(1, lengthAnnotation);	        		
        		Class<?> componentClass = argumentsAnnotation!=null ? argumentsAnnotation.value()[0] : request.clazz.getComponentType();

        		org.simantics.databoard.util.Range[] lengths = null;
        		if (lengthAnnotation!=null) {
            		String[] strs = lengthAnnotation.value();
            		lengths = new org.simantics.databoard.util.Range[strs.length];
            	    for (int i=0; i<strs.length; i++)
            	    	lengths[i] = org.simantics.databoard.util.Range.valueOf(strs[i]);
        			
        		}
	            if (componentClass==null) {
	            	throw new BindingConstructionException("Cannot determine array component type");
	            }
        		
	            ArrayType type = new ArrayType();
	            ArrayBinding binding = null;
	            if (request.clazz.isArray()) {
	        		Class<?> c = request.clazz;
	        		if (c.equals(boolean[].class)) binding = new BooleanArrayBinding(type, null);
	        		else if (c.equals(byte[].class)) binding = new ByteArrayBinding(type, null);
	        		else if (c.equals(double[].class)) binding = new DoubleArrayBinding(type, null);
	        		else if (c.equals(float[].class)) binding = new FloatArrayBinding(type, null);
	        		else if (c.equals(int[].class)) binding = new IntArrayBinding(type, null);
	        		else if (c.equals(long[].class)) binding = new LongArrayBinding(type, null);
	        		else if (c.equals(Object[].class)) binding = new ObjectArrayBinding(type, null);
	        		else binding = new GenericArrayBinding(request.clazz, type, null);		            	
	            } else if (LinkedList.class.isAssignableFrom(request.clazz)) {
           			binding = new LinkedListBinding(type, null);
	            } else if (List.class.isAssignableFrom(request.clazz)) {
           			binding = new ArrayListBinding(type, null);
	            }
				if (binding==null) throw new BindingConstructionException("Cannot bind to "+request.clazz.getSimpleName());
	            
				inprogress.put(request, binding);
				BindingRequest componentRequest = new BindingRequest(componentClass, componentAnnotations);
				binding.componentBinding = construct( componentRequest );
				type.componentType = binding.componentBinding.type();
				inprogress.remove(request);
				
	            // Multi-dimensional Array. Apply range to sub-array-types
	            ArrayType t = type;
	            if (lengths!=null) {
	                for (int i=0; i<lengths.length; i++) {
	                	t.setLength( lengths[i] );
	                    if (i<lengths.length-1)
	                        t = (ArrayType) t.componentType;
	                }
	            }
	            
	    		// Its complete, store to repository
				repository.put(request, binding);			
				return binding;		            
			}
		}
		
		
		
        // Set Type
        if ( Set.class.isAssignableFrom(request.clazz) || request.clazz.equals(TreeSet.class) || request.clazz.equals(HashSet.class) ) {
        	TypeVariable<?>[] classParams = request.clazz.getTypeParameters();
        	
    		Arguments argumentsAnnotation = request.getAnnotation(Arguments.class);
    		Annotation[] componentAnnotations = request.dropAnnotations(1);	        		
    		Class<?> keyClass = argumentsAnnotation!=null ? argumentsAnnotation.value()[0] : request.clazz.getComponentType();	        	
    		if (keyClass==null && classParams.length==1 && classParams[0].getGenericDeclaration() instanceof Class) {
    			keyClass = (Class<?>) classParams[0].getGenericDeclaration();
    		}
    			        
            if (keyClass==null) {
            	throw new BindingConstructionException("Cannot determine key class");
            }

            MapType type = new MapType();
            MapBinding binding;
            
            if (TreeSet.class.isAssignableFrom(request.clazz)) {
            	binding = new TreeSetBinding(type, null);
            } else {
            	binding = new HashSetBinding(type, null);
            }
            
			inprogress.put(request, binding);
			Binding keyBinding = construct( new BindingRequest(keyClass, componentAnnotations) );
			Binding valueBinding = construct( new BindingRequest( void.class ) );
			type.setKeyType(keyBinding.type());
			type.setValueType(valueBinding.type());
			binding.setKeyBinding( keyBinding );
			binding.setValueBinding( valueBinding );
			inprogress.remove(request);	            
			repository.put(request, binding);
            return binding;
        }
		
		
        // Map Type
        if ( Map.class.isAssignableFrom(request.clazz) || request.clazz.equals(TreeMap.class) || request.clazz.equals(HashMap.class) ) {
        	TypeVariable<?>[] classParams = request.clazz.getTypeParameters();
        	
    		Arguments argumentsAnnotation = request.getAnnotation(Arguments.class);
    		Annotation[] componentAnnotations = request.dropAnnotations(2);	        		
    		Class<?> keyClass = argumentsAnnotation!=null ? argumentsAnnotation.value()[0] : request.clazz.getComponentType();	        	
    		Class<?> valueClass = argumentsAnnotation!=null && argumentsAnnotation.value().length>=2 ? argumentsAnnotation.value()[1] : request.clazz.getComponentType();	        	
        	
    		if (keyClass==null && classParams.length>=1 && classParams[0].getGenericDeclaration() instanceof Class) {
    			keyClass = (Class<?>) classParams[0].getGenericDeclaration();
    		}
    		if (valueClass==null && classParams.length>=2 && classParams[1].getGenericDeclaration() instanceof Class) {
    			valueClass = (Class<?>) classParams[0].getGenericDeclaration();
    		}
            if (keyClass==null||valueClass==null) {
            	throw new BindingConstructionException("Cannot determine key and value class");
            }
            BindingRequest keyRequest = new BindingRequest(keyClass, componentAnnotations);
            BindingRequest valueRequest = new BindingRequest(valueClass, componentAnnotations);
            
            MapType type = new MapType();
            MapBinding binding;
            if (TreeMap.class.isAssignableFrom(request.clazz)) {
            	binding = new TreeMapBinding(type, null, null);
            } else {
            	binding = new HashMapBinding(type, null, null);
            }
            
			inprogress.put(request, binding);
			Binding keyBinding = construct( keyRequest );
			Binding valueBinding = construct( valueRequest );
			type.setKeyType(keyBinding.type());
			type.setValueType(valueBinding.type());
			binding.setKeyBinding(keyBinding);
			binding.setValueBinding(valueBinding);
			inprogress.remove(request);
			repository.put(request, binding);
            return binding;
        }
        
        if (request.clazz.isEnum()) {
            Enum<?>[] enums = (Enum[]) request.clazz.getEnumConstants();
            UnionType type = new UnionType();	            
            type.components = new Component[enums.length];
            for (int i=0; i<enums.length; i++) {
                String name = enums[i].name();
                type.components[i] = new Component(name, new RecordType(false));                
            }            	            
    	    EnumClassBinding binding = new EnumClassBinding(type, (Class<Enum<?>>) request.clazz );    	    	            
            repository.put(request, binding);
            return binding;
        }	        
        
        // Union type
        if(request.hasAnnotation(Union.class)) {
	        Union union = request.getAnnotation(Union.class);		        
        	UnionType type = new UnionType();
			UnionClassBinding binding = new UnionClassBinding(type);
            Class<?>[] cases = union.value();
			int count = cases.length; 
            Binding[] componentBindings = new Binding[count];
            type.components = new Component[ count ];
            binding.componentClasses = new Class<?>[ count ];
            binding.setComponentBindings(componentBindings);
            
			inprogress.put(request, binding);	            
            for(int i=0;i<count;++i) {
            	binding.componentClasses[i]= cases[i];
            	Component component = type.components[i] = new Component(cases[i].getSimpleName(), null /* updated later*/);
            	BindingRequest componentRequest = new BindingRequest( cases[i] );
            	Binding componentBinding = componentBindings[i] = construct( componentRequest );
            	component.type = componentBinding.type();	            	
            }	            	           	
			inprogress.remove(request);
			repository.put(request, binding);
            return binding;
        }
        
        // Variant Type
        if (request.clazz == Object.class) {
        	Binding binding = new ObjectVariantBinding(this.repository);
        	repository.put(request, binding);
        	return binding;
        }	        
        if (request.clazz == Void.class || request.clazz == void.class) {
        	Binding binding = VoidBinding.VOID_BINDING;
        	repository.put(request, binding);
        	return binding;
        }	        
        if (request.clazz == Variant.class) {
        	Binding binding = Bindings.VARIANT;
        	repository.put(request, binding);
        	return binding;
        }	        
        if (request.clazz == MutableVariant.class) {
        	Binding binding = Bindings.MUTABLE_VARIANT;
        	repository.put(request, binding);
        	return binding;
        }
        
        // Record type
        {
        	RecordType type = new RecordType();
        	RecordClassBinding binding = new RecordClassBinding(request.clazz, type);	        	
            Field[] fields = getFields(request.clazz);
        	int count = fields.length;
        	Binding[] componentBindings = binding.getComponentBindings();	        	
            Component[] components = new Component[ count ];
            type.referable = request.getAnnotation(Referable.class) != null;
            type.setComponents( components );

			inprogress.put(request, binding);	            
            for(int i=0;i<type.count();++i) {
            	Field field = fields[i];
            	Component c = components[i] = new Component(field.getName(), null /* updated later */);
            	Class<?> fieldClass = field.getType(); 
            	Annotation[] annotations = getFieldAnnotations(field);            	            
            	BindingRequest componentRequest = new BindingRequest(fieldClass, annotations);            	
            	Binding componentBinding = componentBindings[i] = construct( componentRequest );
            	c.type = componentBinding.type();
            }
            
			inprogress.remove(request);
			repository.put(request, binding);
            return binding;
        }
	}
	
	public Binding construct(BindingRequest request) 
	throws BindingConstructionException
	{
		if (failures.containsKey(request)) throw failures.get(request);
		if (inprogress.containsKey(request)) return inprogress.get(request);
		if (repository.containsKey(request)) return repository.get(request);
		
		// Start construction
		try {			
			Binding binding = doConstruct(request);
			repository.put(request, binding);
			return binding;
		} catch (RangeException e) {
			inprogress.remove( request );
			BindingConstructionException bce = new BindingConstructionException( e ); 
			failures.put(request, bce);
			throw bce;
		} catch (BindingConstructionException e) {
			inprogress.remove( request );
			failures.put(request, e);
			throw e;
		}		
		
	}

	/**
	 * Get a binding to a Java Class. Type details can be modified with annotations.
	 * Please see the package org.simantics.databoard.annotations. 
	 * <p>
	 *  
	 * The whether the result binding is a completely mutable or not depends on the
	 * provided classes. For instance, fields such as Boolean, Integer, Long
	 * are not mutable, instead MutableBoolean, MutableInteger and MutableLong are.
	 * The length of Object[] is not mutable, but length of List<Object> is. <p>
	 * 
	 * Note, results are stored with strong reference into the repository assigned
	 * to this factory.<p> 
	 * 
	 * @see ReflectionBindingFactory  
	 * @param clazz
	 * @return binding
	 * @throws BindingConstructionException
	 */
    @SuppressWarnings("unchecked")
	public <T extends Binding> T getBinding(Class<?> clazz)
    throws BindingConstructionException
    {
    	return (T) construct( new BindingRequest(clazz) );
    }
    
	/**
	 * Get a binding to a Java Class. Use this method to acquire class 
	 * parameters for a generics class. <p>
	 * 
	 * Example 1: 
	 * 
	 *    Binding binding = getBinding(Map.class, String.class, Integer.class);
	 *    Map<String, Integer> map = (Map<String, Integer>) binding.createDefault();
	 *    
	 * Example 2:
	 *    
     *  Binding d = getBinding(List.class, Integer.class);
	 *  List<Integer> list = (List<Integer>) d.createRandom(5);
	 *    
	 * Example 3:
	 *    
     *  Binding d = getBinding(List.class, List.class, Integer.class);
	 *  List<List<Integer>> list = (List<List<Integer>>) d.createRandom(5);
	 * 
	 * @see ReflectionBindingFactory  
	 * @param clazz
	 * @return binding
	 * @throws BindingConstructionException
	 */
    @SuppressWarnings("unchecked")
	public <T extends Binding> T getBinding(Class<?> clazz, Class<?>...parameters)
    throws BindingConstructionException
    {
    	Arguments args = new ArgumentImpl(parameters);
    	BindingRequest request = new BindingRequest( clazz, args );
    	return (T) construct( request );
    }    
	
	public Binding getBinding(BindingRequest request) throws BindingConstructionException {		
		return construct(request);
	}
	
	public Binding getBindingUnchecked(BindingRequest request) throws RuntimeBindingConstructionException {
		try {
			return construct(request);
		} catch (BindingConstructionException e) {
			throw new RuntimeBindingConstructionException(e);
		}
	}	
	
	
    static Annotation[] getFieldAnnotations(Field field) 
    {
    	Class<?> fieldClass = field.getType(); 
    	Annotation[] annotations = field.getAnnotations().clone();
    	Class<?>[] parameterClasses = getParameterClasses(field);
    	
    	if (Set.class.isAssignableFrom(fieldClass) && parameterClasses!=null &&parameterClasses.length==1) {
    		Annotation[] a2 = new Annotation[annotations.length+1];
    		System.arraycopy(annotations, 0, a2, 0, annotations.length);
    		
    		Class<?> keyType = parameterClasses[0];
    		a2[annotations.length] = new ArgumentImpl(keyType);						
    		annotations = a2;
    	}    	
    	if (Map.class.isAssignableFrom(fieldClass) && parameterClasses!=null && parameterClasses.length==2) {
    		Annotation[] a2 = new Annotation[annotations.length+1];
    		System.arraycopy(annotations, 0, a2, 0, annotations.length);
    		
    		Class<?> keyType = parameterClasses[0];
    		Class<?> valueType = parameterClasses[1];
    		a2[annotations.length] = new ArgumentImpl(keyType, valueType);						
    		annotations = a2;
    	}    	
    	if (List.class.isAssignableFrom(fieldClass) && parameterClasses!=null && parameterClasses.length==1) {
    		Annotation[] a2 = new Annotation[annotations.length+1];
    		System.arraycopy(annotations, 0, a2, 0, annotations.length);
    		Class<?> componentType = parameterClasses[0];
    		a2[annotations.length] = new ArgumentImpl(componentType);
    		annotations = a2;
    	}    	
    	
    	return annotations;
    }
    
    
    /**
     * Get all actual parameters classes, omits generic type definitions
     * 
     * @param f
     * @return
     */
    @SuppressWarnings("unchecked")
	static Class[] getParameterClasses(Field f) {
		Type t = f.getGenericType();
		if (t==null || t instanceof ParameterizedType==false) return new Class[0];		
		ParameterizedType p = (ParameterizedType) t;
		
		int count = 0;
		for (Type x : p.getActualTypeArguments() )
			if (x instanceof Class) count++;
		
		Class[] result = new Class[count];
		int i = 0; 
		for (Type x : p.getActualTypeArguments() )
			if (x instanceof Class)
				result[i++] = (Class) x;
		return result;
    }

    /**
     * Get all actual parameters types, incl. classes and generic types
     * 
     * @param f
     * @return
     */
	static Type[] getParameterTypes(Field f) {
		Type t = f.getGenericType();
		if (t==null || t instanceof ParameterizedType==false) return new Class[0];		
		ParameterizedType p = (ParameterizedType) t;
		return p.getActualTypeArguments();
    }
    
	/**
	 * Get all, including inherited, public, protected and default 
	 * non-transient, non-static, non-final fields. All returned fields have access.
	 * 
	 * @param clazz
	 * @return
	 */
	static Field[] getFields(Class<?> clazz)
	{
		// TODO Sort as order is not guaranteed
		// Sort priority 1-class, 2-modifiers, 3-name, 4-signature
		Field[] fields = getAllFields(clazz);
		
		ArrayList<Field> result = new ArrayList<Field>(fields.length);
		for (Field f : fields) {
			if (Modifier.isStatic(f.getModifiers())) continue;
//			if (Modifier.isFinal(f.getModifiers())) continue;
            if (Modifier.isTransient(f.getModifiers())) continue;           
            if (Modifier.isVolatile(f.getModifiers())) continue;           
			f.setAccessible(true);
			result.add(f);
		}
		return result.toArray( new Field[result.size()] );
	}
	
	// Get all fields except fields of Throwable
	static Field[] getAllFields(Class<?> clazz)
	{
		LinkedList<Class<?>> classes = new LinkedList<Class<?>>();
		while (clazz!=null) {
			classes.addFirst(clazz);
			clazz = clazz.getSuperclass();
			if (clazz==Throwable.class) clazz=null;
		}
		
		ArrayList<Field> result = new ArrayList<Field>();
		for (Class<?> _class : classes) {
			_getAllFields(_class, result);
		}
		
		return result.toArray(new Field[result.size()]);
	}
	
	public static void _getAllFields(Class<?> clazz, Collection<Field> result)
	{
		for (Field m : clazz.getDeclaredFields())
			result.add(m);
	}	    
	

}
