package org.simantics.databoard.channel;

import java.io.IOException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.simantics.databoard.binding.Binding;

/**
 *
 * @author Toni Kalajainen <toni.kalajainen@iki.fi>
 */
public interface AsyncRequest {

	/**
	 * Get request status
	 * 
	 * @return status
	 */
	Status getStatus(); 	
	enum Status {Waiting, Succeed, Failed}	
	
	/**
	 * Get result if available. 
	 * 
	 * @param binding
	 * @return result or <tt>null</tt>
	 * @throws RequestException
	 */
	Object getResult(Binding binding) throws RequestException;
	
	/**
	 * Get error if available.
	 * 
	 * Cause is (typically) one of these:
	 *  o Communication problems {@link IOException}
	 *  o Timeout {@link TimeoutException}
	 *  o Service handler problem {@link ServiceException} 	 * 
	 * 
	 * @return error or <tt>null</tt>
	 */
	Exception getError(); 	
	
	/**
	 * Wait for result until default timeout. 
	 * Default timeout is configured to the channel. 
	 * If timeout occurs TimeoutException is thrown wrapped in RequestException.
	 * 
	 * @param binding
	 * @return the result object
	 * @throws RequestException
	 */
	Object waitForResult(Binding binding) throws RequestException;
	
	/**
	 * Wait for result or break after until timeout. 
	 * If timeout occurs TimeoutException is thrown wrapped in RequestException.
	 * 
	 * @param binding
	 * @param timeout
	 * @param unit
	 * @return result
	 * @throws RequestException
	 */
	Object waitForResult(Binding binding, long timeout, TimeUnit unit) throws RequestException;

	/**
	 * Set a listener. If the result is already available, the listener
	 * is invoked from the current thread.  
	 * 
	 * Otherwise, the notification is placed from the thread that sets
	 * the result. The listener implementation may not create any locks
	 * during the handling of the result, also it is good policy not do
	 * any long term operations in the listener implementation.  
	 *  
	 * @param listener (listener may not block) or null to remove listener
	 */
	void setListener(RequestListener listener);
	
	interface RequestListener {
		/**
		 * Request completed, the result is available
		 * 
		 * @param result the result 
		 */
		void onCompleted(Object result);
		
		/**
		 * There was an error in processing the request
		 * 
		 * @param error the error
		 */
		void onError(ServiceException error);
		
	}
	
}
