/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.datasource;

import java.util.Collection;
import java.util.Map;
import java.util.Random;

import org.simantics.databoard.Bindings;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.impl.DoubleBindingDefault;
import org.simantics.databoard.binding.mutable.MutableVariant;
import org.simantics.databoard.binding.mutable.Variant;

/**
 * Test Datasource provides 5 example values, one of which is time.
 * 
 * The implementation is an accessor that follows Datasource contract.
 * 
 * New values are spawned on #step().
 * 
 * Variables:
 *   SINE    - Double(range=[-1..1]) 
 *   RAMP    - Double(range=[0..100])
 *   ARRAY   - int[3]
 *   VECTOR  - { x: Double; y: Double; z: Double }
 *   TIME    - Double(unit="s")
 *   FOLDER  - no value, has child RAMP
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class ExampleDatasource extends StepwiseDatasource {
	
	Sine sine;
	Time time;
	Ramp ramp;
	Array array;
	Vector vector;
	Folder folder;
	
	public ExampleDatasource() {
		super();
		addNode( sine = new Sine() );
		addNode( time = new Time() );
		addNode( ramp = new Ramp() );
		addNode( array = new Array() );
		addNode( vector = new Vector() );
		addRoot( folder = new Folder(sine.id, time.id, ramp.id, array.id, vector.id) );
	}

	@Override
	public void doStep() throws DatasourceException {
		time.value += 0.100;
	}
	
	class Sine implements Node {

		Variant id = Variant.ofInstance("SINE");
		
		@Override
		public Variant getId() {
			return id;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
		}

		@Override
		public boolean getValue(MutableVariant value) {
			double t = time.value;
			Double y = Math.sin(t);			
			value.setValue( Bindings.DOUBLE, y);
			return true;
		}

		@Override
		public String getLabel(String locale) {
			if (locale.equals("fi")) return "Sini-kyr";
			if (locale.equals("en")) return "Sine-wave";
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			locales.put("fi", "Sini-kyr");
			locales.put("en", "Sine-wave");
		}

		@Override
		public boolean setValue(Variant value) {
			return false;
		}
		
	}
	
	class Ramp implements Node {

		MutableVariant id = MutableVariant.ofInstance("RAMP");
		
		@Override
		public Variant getId() {
			return id;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
		}

		@Override
		public boolean getValue(MutableVariant value) {
			double remainder = time.value % 1;
			double y = remainder < 0.5 ? 0.0 : 1.0;
			value.setValue( Bindings.DOUBLE, y);
			return true;
		}

		@Override
		public String getLabel(String locale) {
			if (locale.equals("fi")) return "Kanttiaalto";
			if (locale.equals("en")) return "Ramp";
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			locales.put("fi", "Kanttiaalto");
			locales.put("en", "Ramp");
		}

		@Override
		public boolean setValue(Variant value) {
			return false;
		}
		
	}
	
	class Array implements Node {

		MutableVariant id = MutableVariant.ofInstance("ARRAY");
		int[] value = new int[] {0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
		Binding binding = Bindings.getBindingUnchecked( int[].class );
		Random r = new Random();
		
		@Override
		public Variant getId() {
			return id;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
		}

		@Override
		public boolean getValue(MutableVariant value) {
			int result[] = new int[10];
			for (int i=0; i<10; i++) 
				result[i] = r.nextInt();
			value.setValue(binding, this.value);
			return true;
		}

		@Override
		public String getLabel(String locale) {
			if (locale.equals("fi")) return "Taulukko";
			if (locale.equals("en")) return "Array";
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			locales.put("fi", "Taulukko");
			locales.put("en", "Array");
		}

		@Override
		public boolean setValue(Variant value) {
			return false;
		}
		
	}
	
	public static class VectorRecord {
		public double x, y, z; 
	}
	
	class Vector implements Node {

		Variant id = Variant.ofInstance("VECTOR");
		Binding binding = Bindings.getBindingUnchecked( VectorRecord.class );
		Random r = new Random();			
		
		@Override
		public Variant getId() {
			return id;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
		}

		@Override
		public boolean getValue(MutableVariant value) {
			VectorRecord vr = new VectorRecord();
//			vr.x = r.nextDouble();
			vr.y = r.nextDouble();
//			vr.z = r.nextDouble();
			value.setValue(binding, vr);
			return true;
		}

		@Override
		public String getLabel(String locale) {
			if (locale.equals("fi")) return "Vektori";
			if (locale.equals("en")) return "Vector";
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			locales.put("fi", "Vektori");
			locales.put("en", "Vector");
		}

		@Override
		public boolean setValue(Variant value) {
			return false;
		}
		
	}
	

	
	class Time implements Node {

		Variant id = Variant.ofInstance("TIME");
		Double value = 0.0;
		
		@Override
		public Variant getId() {
			return id;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
		}

		@Override
		public boolean getValue(MutableVariant value) {
			value.setValue( Bindings.DOUBLE, this.value);
			return true;
		}

		@Override
		public String getLabel(String locale) {
			if (locale.equals("fi")) return "Aika";
			if (locale.equals("en")) return "Time";
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			locales.put("fi", "Aika");
			locales.put("en", "Time");
		}

		@Override
		public boolean setValue(Variant value) throws DatasourceException {
			try {
				this.value = (Double) value.getValue( Bindings.DOUBLE );
			} catch (AdaptException e) {
				throw new DatasourceException( e );
			}
			return true;
		}
		
	}


	class Folder implements Node {

		Variant id = Variant.ofInstance("FOLDER");
		Variant children[];
		
		public Folder(Variant...content) {
			children = content;
		}
		
		@Override
		public Variant getId() {
			return id;
		}

		@Override
		public void getChildren(Collection<Variant> children) {
			for (Variant id : this.children)
				children.add( id );
		}

		@Override
		public boolean getValue(MutableVariant value) {
			return false;
		}

		@Override
		public String getLabel(String locale) {
			if (locale.equals("fi")) return "Kansio";
			if (locale.equals("en")) return "Folder";
			return null;
		}

		@Override
		public void getLabels(Map<String, String> locales) {
			locales.put("fi", "Kansio");
			locales.put("en", "Folder");
		}

		@Override
		public boolean setValue(Variant value) throws DatasourceException {
			return false;
		}
		
	}
	
}

