/*******************************************************************************
 * Copyright (c) 2010- Association for Decentralized Information Management in
 * Industry THTH ry.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *    VTT Technical Research Centre of Finland - initial API and implementation
 *******************************************************************************/
package org.simantics.databoard.datasource;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.simantics.databoard.Accessors;
import org.simantics.databoard.Bindings;
import org.simantics.databoard.accessor.Accessor;
import org.simantics.databoard.accessor.MapAccessor;
import org.simantics.databoard.accessor.RecordAccessor;
import org.simantics.databoard.accessor.emit.CollectingEmitter;
import org.simantics.databoard.accessor.emit.ImmediateEmitter;
import org.simantics.databoard.accessor.error.AccessorConstructionException;
import org.simantics.databoard.accessor.error.AccessorException;
import org.simantics.databoard.accessor.error.ReferenceException;
import org.simantics.databoard.accessor.error.WriteAccessException;
import org.simantics.databoard.accessor.event.Event;
import org.simantics.databoard.accessor.event.MapEntryAdded;
import org.simantics.databoard.accessor.event.MapEntryRemoved;
import org.simantics.databoard.accessor.event.ValueAssigned;
import org.simantics.databoard.accessor.impl.AccessorParams;
import org.simantics.databoard.accessor.impl.CompositeRecord;
import org.simantics.databoard.accessor.interestset.InterestSet;
import org.simantics.databoard.accessor.interestset.MapInterestSet;
import org.simantics.databoard.accessor.interestset.OptionalInterestSet;
import org.simantics.databoard.accessor.interestset.RecordInterestSet;
import org.simantics.databoard.accessor.interestset.VariantInterestSet;
import org.simantics.databoard.accessor.java.JavaObject;
import org.simantics.databoard.accessor.java.JavaOptional;
import org.simantics.databoard.accessor.java.JavaVariant;
import org.simantics.databoard.accessor.reference.ChildReference;
import org.simantics.databoard.accessor.reference.KeyReference;
import org.simantics.databoard.accessor.reference.LabelReference;
import org.simantics.databoard.adapter.AdaptException;
import org.simantics.databoard.adapter.Adapter;
import org.simantics.databoard.adapter.AdapterConstructionException;
import org.simantics.databoard.binding.ArrayBinding;
import org.simantics.databoard.binding.Binding;
import org.simantics.databoard.binding.MapBinding;
import org.simantics.databoard.binding.RecordBinding;
import org.simantics.databoard.binding.error.BindingException;
import org.simantics.databoard.binding.impl.ArrayListBinding;
import org.simantics.databoard.binding.impl.StringBindingDefault;
import org.simantics.databoard.binding.impl.StringVariantBinding;
import org.simantics.databoard.binding.impl.TreeMapBinding;
import org.simantics.databoard.binding.mutable.ContainerOptionalBinding;
import org.simantics.databoard.binding.mutable.ImmutableVariantBinding;
import org.simantics.databoard.binding.mutable.MutableVariant;
import org.simantics.databoard.binding.mutable.MutableVariantBinding;
import org.simantics.databoard.binding.mutable.ValueContainer;
import org.simantics.databoard.binding.mutable.Variant;
import org.simantics.databoard.datasource.IStepwiseDatasource.DatasourceListener;
import org.simantics.databoard.type.MapType;

/**
 * Adapt Datasource into a read-only accessor.
 * 
 * This implementation does not support modifications to the datamodel and 
 * labels of the data source.<p> 
 * 
 * Concurrent usage model is the following:
 * 
 *  On pre-step, Values are copied from accessor to datasource. Accessor is read&write locked.
 *  During step, Accessor is write locked.
 *  On post-step, Values are copied from datasource to accessor. Accessor is read&write locked.
 *  Outside step, accessor is readable and writable to any thread. 
 *
 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
 */
public class StepwiseDatasourceAccessor extends CompositeRecord implements DatasourceListener {

	// Bindings for node fields
	static Binding MUTABLE_VARIANT_BINDING = Bindings.MUTABLE_VARIANT;
	static Binding IMMUTABLE_VARIANT_BINDING = Bindings.VARIANT;
	static Binding LOCALIZEDTEXT_BINDING = new TreeMapBinding( Bindings.STRING, Bindings.STRING );
	static Binding VARIANT_ARRAY_BINDING = new ArrayListBinding( Bindings.MUTABLE_VARIANT );
	static Binding OPTIONAL_VARIANT_BINDING = new ContainerOptionalBinding( Bindings.MUTABLE_VARIANT );
	static Binding NODE_BINDING = Bindings.getBindingUnchecked( NodeRecord.class );
	static MapType NODES_TYPE;
	
	IStepwiseDatasource datasource;
	NodesMapAccessor nodesAccessor;
	DatasourceListener listener;
	Lock readLock, writeLock;
	boolean inStep = false;
	/** A list of nodes that were modified outside step. Modifications are applied  */
	Map<Variant, JavaVariant> externalModifications = new HashMap<Variant, JavaVariant>();
	
	public StepwiseDatasourceAccessor(IStepwiseDatasource datasource) {
		this.datasource = datasource;
		ReentrantReadWriteLock llock = new ReentrantReadWriteLock(false);
		readLock = llock.readLock();
		writeLock = llock.writeLock();
		params = new AccessorParams(readLock, writeLock, ImmediateEmitter.INSTANCE, false);
		
		addField("nodes", nodesAccessor = new NodesMapAccessor());
		
		// Collect modifications that occur outside step (from outsiders)
		// Put them to externalModifications map
		Listener nodeModificationListener = new Listener() {
			@Override
			public void onEvents(Collection<Event> events) {
				if (inStep) return;
								
				for (Event e : events) {
					try {
						if (e.reference instanceof KeyReference) {
							KeyReference kr = (KeyReference) e.reference;
							Variant nodeId = (Variant) kr.key.getValue( Bindings.VARIANT );
							NodeAccessor na = nodesAccessor.getValueAccessor( Bindings.VARIANT, nodeId);						
							JavaVariant va = (JavaVariant) na.valueAccessor.getComponentAccessor();
							externalModifications.put(nodeId, va);													
						}
					} catch (AccessorConstructionException e1) {
						throw new RuntimeException(e1);
					} catch (AdaptException e2) {
						throw new RuntimeException(e2);
					}
				}
			}
		};
		VariantInterestSet isVariant = VariantInterestSet.MONITOR_EVERYTHING;
		OptionalInterestSet isValue = new OptionalInterestSet(false, false, isVariant); 
		RecordInterestSet isNode = new RecordInterestSet(false, null, false, null, new InterestSet[] {null, null, null, isValue});
		MapInterestSet isNodes = new MapInterestSet(false, null, false, null, isNode, null);
		// Attach listener to nodes - Map(Variant, Variant)
		try {
			nodesAccessor.addListener(nodeModificationListener, isNodes, null);
		} catch (AccessorException e) {
			throw new RuntimeException(e);
		}
		
		datasource.addListener(this);
	}
	
	@Override
	public void onPreStep() {
		writeLock.lock();
		try {
			inStep = true;
			// Transfer written values from accessor to datasource
			
			// How do we know which variables were changed?
			monitoredNodes.clear();
//			getActiveNodes(activeNodes);
			
			for (Entry<Variant, JavaVariant> e : externalModifications.entrySet()) {
				try {
					Variant nodeId = e.getKey();
					JavaVariant value = e.getValue();
					datasource.setValue(nodeId, (Variant) value.getValue( Bindings.VARIANT ));
				} catch (DatasourceException e1) {
					e1.printStackTrace();
				} catch (AccessorException e2) {
					e2.printStackTrace();
				}
			}
		} finally {
			externalModifications.clear();
//			readLock.lock();
//			writeLock.unlock();
		}		
	}
		
	HashSet<NodeAccessor> monitoredNodes = new HashSet<NodeAccessor>();
	@Override
	public void onPostStep() {
		// Transfer values from datasource to accessor
		try {
			monitoredNodes.clear();
			getMonitoredNodes(monitoredNodes);
			// Collect events into a single set for each listener 
			CollectingEmitter mechanism = new CollectingEmitter(); 
			params.emitter = mechanism;
			
			MutableVariant value = new MutableVariant();
			for (NodeAccessor na : monitoredNodes) {
				Variant nodeId = na.nodeId;
				boolean hasValue = datasource.getValue(nodeId, value);
				
				if (!hasValue) {
					na.valueAccessor.setNoValue();
				} else {
					na.valueAccessor.setComponentValue( Bindings.MUTABLE_VARIANT, value);
				}
			}
			
			// Forward events to outsider listeners
			mechanism.forwardEvents();
		} catch (DatasourceException e) {
			e.printStackTrace();
		} catch (AccessorException e) {
			e.printStackTrace();
		} catch (AccessorConstructionException e) {
			e.printStackTrace();
		} finally {
			inStep = false;
			params.emitter = ImmediateEmitter.INSTANCE;
//			readLock.unlock();
			writeLock.unlock();
		}
	}
	
	public void dispose() {
		datasource.removeListener( listener );
	}
	
	/**
	 * Get all nodes that are being monitored by accessor's users. This is a list of non-garbage
	 * collected accessors.
	 * 
	 * @param result list to be filled with nodes
	 * @throws AccessorConstructionException 
	 */
	void getMonitoredNodes(Set<NodeAccessor> result) throws AccessorConstructionException {
		for (java.lang.ref.Reference<NodeAccessor> ref : nodesAccessor.children.values()) {
			NodeAccessor node = ref.get();
			if (node!=null)	result.add( node );
		}
		nodesAccessor.getActiveNodes(result);		
	}
	
	
	static {
		NODES_TYPE = new MapType( MUTABLE_VARIANT_BINDING.type(), NODE_BINDING.type() );
	}
	
	class NodeAccessor extends CompositeRecord {
		
		Variant nodeId;
		Variant wrappedNodeId;
		Accessor idAccessor;
		Accessor labelsAccessor;
		Accessor childrenAccessor;
		JavaOptional valueAccessor;
		
		public NodeAccessor(NodesMapAccessor parent, Variant nodeId) throws AccessorConstructionException, DatasourceException {
			super(parent);
			this.nodeId = nodeId;
			this.wrappedNodeId = new Variant( IMMUTABLE_VARIANT_BINDING, nodeId );
			this.params = StepwiseDatasourceAccessor.this.params;
			
			// Identifier
			idAccessor = JavaObject.createAccessor( this, MUTABLE_VARIANT_BINDING, nodeId, params );
			
			// Read labels
			TreeMap<String, String> labels = new TreeMap<String, String>();
			datasource.getLabels(nodeId, labels);
			labelsAccessor = JavaObject.createAccessor( this, LOCALIZEDTEXT_BINDING, labels, params );
			
			// Read children
			ArrayList<Variant> children = new ArrayList<Variant>();
			datasource.getChildren(nodeId, children);
			childrenAccessor = JavaObject.createAccessor( this, VARIANT_ARRAY_BINDING, children, params);
			
			// Value Accessor
//			ValueContainer container = new ValueContainer( nodeId );
//			MutableVariant value = new MutableVariant();
//			// Read value
//			if (datasource.getValue(nodeId, value)) container.setValue( value );
//			Binding valueBinding = new OptionalBindingDefault( new ValueBinding(datasource) );
//			valueAccessor = (JavaOptional) JavaObject.createAccessor( this, valueBinding, nodeId);

			ValueContainer container = new ValueContainer();
			MutableVariant value = new MutableVariant();
			// Read value
			if (datasource.getValue(nodeId, value)) container.setValue( value );
			valueAccessor = (JavaOptional) JavaObject.createSubAccessor( this, OPTIONAL_VARIANT_BINDING, container, params );
			
			addField("id", idAccessor);
			addField("labels", labelsAccessor);
			addField("children", childrenAccessor);
			addField("value", valueAccessor);			
		}
		
		void setValue(MutableVariant newValue) {
			try {
				valueAccessor.setValue(MUTABLE_VARIANT_BINDING, newValue);
			} catch (AccessorException e) {
				throw new RuntimeException( e );
			}
		}
	}
	
//	/**
//	 * Binds nodeId to a node value field. 
//	 * If node is folder, there is no value, otherwise there is.
//	 *
//	 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
//	 */
//	static class NodeBinding extends OptionalBinding {
//
//		IStepwiseDatasource datasource;
//		
//		public NodeBinding(IStepwiseDatasource datasource) {
//			super(new ValueBinding(datasource));
//			this.datasource = datasource;
//		}
//
//		@Override
//		public Object createNoValue() throws BindingException {
//			throw new BindingException("Cannot create new node");
//		}
//
//		@Override
//		public Object createValue(Object value) throws BindingException {
//			throw new BindingException("Cannot create new node");
//		}
//
//		@Override
//		public boolean hasValue(Object arg) throws BindingException {
//			try {
//				Variant nodeId = (Variant) arg;
//				MutableVariant value = new MutableVariant();
//				boolean ok = datasource.getValue(nodeId, value);
//				return ok;
//			} catch (DatasourceException e) {
//				throw new BindingException(e);
//			}
//		}
//
//		@Override
//		public Object getValue(Object arg) throws BindingException {			
//			return arg;
//		}
//
//		@Override
//		public void setValue(Object optional, Object componentValue)
//				throws BindingException {
//			throw new BindingException("Cannot remove value");
////			ValueBinding vb = (ValueBinding) getComponentBinding();
////			vb.setContent(optional, contentBinding, content)
//		}
//
//		@Override
//		public void setNoValue(Object optional) throws BindingException {
//			throw new BindingException("Cannot remove value");
//		}
//
//		@Override
//		public boolean isInstance(Object obj) {
//			return obj instanceof Variant;
//		}
//		
//	}
//	
//	/**
//	 * Binds nodeId to the value of the node
//	 *
//	 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
//	 */
//	static class ValueBinding extends VariantBinding {
//		
//		IStepwiseDatasource datasource;
//		
//		public ValueBinding(IStepwiseDatasource datasource) {
//			this.datasource = datasource;
//		}
//
//		@Override
//		public Object getContent(Object variant, Binding contentBinding) throws BindingException {
//			try {
//				Variant nodeId = (Variant) variant;
//				MutableVariant result = new MutableVariant();
//				boolean ok = datasource.getValue(nodeId, result);
//				if (!ok) throw new BindingException("Could not read value");
//				return result.getValue(contentBinding);
//			} catch (DatasourceException e) {
//				throw new BindingException( e );
//			} catch (AdaptException e) {
//				throw new BindingException( e );
//			}
//		}
//
//		@Override
//		public Object getContent(Object variant) throws BindingException {
//			try {
//				Variant nodeId = (Variant) variant;
//				MutableVariant result = new MutableVariant();
//				boolean ok = datasource.getValue(nodeId, result);
//				if (!ok) throw new BindingException("Could not read value");
//				return result;
//			} catch (DatasourceException e) {
//				throw new BindingException( e );
//			}
//		}
//
//		@Override
//		public DataType getContentType(Object variant) throws BindingException {
//			try {
//				Variant nodeId = (Variant) variant;
//				return datasource.getNodeType(nodeId);
//			} catch (DatasourceException e) {
//				throw new BindingException( e );
//			}
//		}
//
//		@Override
//		public Binding getContentBinding(Object variant) throws BindingException {
//			try {
//				Variant nodeId = (Variant) variant;
//				MutableVariant result = new MutableVariant();
//				boolean ok = datasource.getValue(nodeId, result);
//				if (!ok) throw new BindingException("Could not read value");
//				return result.getBinding();
//			} catch (DatasourceException e) {
//				throw new BindingException( e );
//			}
//		}
//
//		@Override
//		public Object create(Binding contentBinding, Object content) throws BindingException {
//			throw new BindingException("Cannot add node");
//		}
//
//		@Override
//		public void setContent(Object variant, Binding contentBinding, Object content) throws BindingException {
//			try {
//				Variant nodeId = (Variant) variant;
//				boolean ok = datasource.setValue(nodeId, new Variant(contentBinding, content));
//				if (!ok) throw new BindingException("Could not write value");
//			} catch (DatasourceException e) {
//				throw new BindingException( e );
//			}
//		}
//
//		@Override
//		public boolean isInstance(Object obj) {
//			return obj instanceof Variant;
//		}
//
//		@Override
//		public void assertInstaceIsValid(Object obj, Set<Object> validInstances) throws BindingException {
//			if (obj instanceof Variant==false) throw new BindingException("not nodeid");
//			Variant nodeId = (Variant) obj;
//			if (!datasource.hasNode(nodeId))
//				throw new BindingException("node not found");
//			if (validInstances!=null) validInstances.add(nodeId);
//		}
//		
//	}
	
	/**
	 * Accessor to datasource nodes. Nodes are read from a IStepwiseDatasource.
	 * 
	 * Sub-accessor for a node is instantiated on demand and stored with a soft
	 * reference.
	 *
	 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
	 */
	class NodesMapAccessor implements MapAccessor {
		/** Accessors to children */
		TreeMap<Variant, java.lang.ref.Reference<NodeAccessor>> children = new TreeMap<Variant, java.lang.ref.Reference<NodeAccessor>>( IMMUTABLE_VARIANT_BINDING );

		TreeSet<Variant> ids;
		
		/** Listeners */
		ListenerEntryEx listeners = null;
		
		@Override
		public MapType type() {
			return NODES_TYPE;
		}
		
		TreeSet<Variant> getIds() {
			if (ids==null) { 
				ids = new TreeSet<Variant>( IMMUTABLE_VARIANT_BINDING );
				datasource.getAll(ids);
			}
			return ids;
		}
		
		void getActiveNodes(Set<NodeAccessor> activeNodes) throws AccessorConstructionException {
			// Add all to which there is a listener
//			ListenerEntryEx le = nodesAccessor.listeners;
//			if (le!=null) {
//				
//				while (le!=null) {				
//					MapInterestSet is = le.getInterestSet();				
//					
//					if (le.path instanceof KeyReference) {
//						KeyReference kr = (KeyReference) le.path;
//						NodeAccessor na = getValueAccessor( kr.key.getBinding(), kr.key.getValue() );
//						activeNodes.add( na );
//					}
//					// Next listener
//					le = le.next;
//				}
//			}
			
		}
		
		/**
		 * Get existing sub accessor
		 * @param index
		 * @return sub-accessor or <code>null</code>
		 */
		NodeAccessor getExistingAccessor(Variant nodeId)
		{
			java.lang.ref.Reference<NodeAccessor> ref = children.get(nodeId);
			if (ref==null) return null;
			NodeAccessor result = ref.get();
			if (result==null) {
				children.remove(nodeId);
				return null;
			}
			return result;
		}	
		
		@Override
		public <T extends Accessor> T getComponent(ChildReference reference)
				throws AccessorConstructionException {
			if (reference==null) return (T) this;
			if (reference instanceof LabelReference) {
				LabelReference lr = (LabelReference) reference;
				try {
					Variant value = (Variant) Bindings.adapt(lr.label, Bindings.STRING, IMMUTABLE_VARIANT_BINDING);
					Accessor result = (T) getValueAccessor(MUTABLE_VARIANT_BINDING, value);
					if (reference.getChildReference() != null)
						result = result.getComponent(reference.getChildReference());
					return (T) result;			
				} catch (AdaptException e) {
					throw new ReferenceException(e);				
				}			
			} else if (reference instanceof KeyReference) {
				KeyReference ref = (KeyReference) reference;
				Accessor result = getValueAccessor(MUTABLE_VARIANT_BINDING, ref.key);
				if (reference.getChildReference() != null)
					result = result.getComponent(reference.getChildReference());
				return (T) result;			
			} 
			throw new ReferenceException(reference.getClass().getName()+" is not a reference of a map");	
		}

		@Override
		public <T extends Accessor> T getValueAccessor(Binding keyBinding,
				Object key) throws AccessorConstructionException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, IMMUTABLE_VARIANT_BINDING);
				if (!datasource.hasNode(nodeId)) throw new AccessorConstructionException("Map doesn't contain the requested element");				
				
				NodeAccessor sa = getExistingAccessor(nodeId);
				if (sa!=null) return (T) sa;
				
				// Create NodeAccessor
				sa = new NodeAccessor(this, nodeId);
				children.put(nodeId, new WeakReference<NodeAccessor>(sa));
				
				// Add component interest sets
				ListenerEntryEx le = listeners;
				if (le!=null) {
					
					while (le!=null) {				
						MapInterestSet is = le.getInterestSet();
	
						// Generic element interest
						InterestSet gis = is.getComponentInterest(); 
						if (gis != null) {
							try {							
								ChildReference childPath = ChildReference.concatenate(le.path, new KeyReference( sa.wrappedNodeId ) );
								sa.addListener(le.listener, gis, childPath);
							} catch (AccessorException e) {
								throw new AccessorConstructionException(e);
							}
						}
							
						// Specific element interest
						InterestSet cis = is.getComponentInterest( sa.wrappedNodeId ); 
						if (cis != null) {
							try {
								ChildReference childPath = ChildReference.concatenate(le.path, new KeyReference( sa.wrappedNodeId ) );
								sa.addListener(le.listener, cis, childPath);
							} catch (AccessorException e) {
								throw new AccessorConstructionException(e);
							}
						}
						
						// Next listener
						le = le.next;
					}
				}
				
				return (T) sa;
			} catch (AdaptException e) {
				throw new AccessorConstructionException(e);
			} catch (DatasourceException e) {
				throw new AccessorConstructionException(e);
			}	
		}
		
		@Override
		public void addListener(Listener listener, InterestSet interestSet, ChildReference path) throws AccessorException {
			ListenerEntryEx l = ListenerEntryEx.link(listeners, listener, interestSet, path);
			listeners = l;
			MapInterestSet is = (MapInterestSet) interestSet;
			
			// Apply listener to existing sub-accessors
			InterestSet cis = is.getComponentInterest();
			if (cis!=null) {
				for (Variant nodeId : children.keySet()) {
					NodeAccessor sa = getExistingAccessor(nodeId);
					if (sa==null) continue;
					ChildReference childPath = ChildReference.concatenate( path, new KeyReference(sa.wrappedNodeId) );
					sa.addListener(listener, cis, childPath);				
				}
			} 
			
			if (is.componentInterests!=null) {
				try {
					for (Entry<Variant,InterestSet> e : is.componentInterests.entrySet()) {
						cis = e.getValue();
						Variant nodeId = e.getKey();
						NodeAccessor sa = getValueAccessor(Bindings.VARIANT, nodeId);
						ChildReference childPath = ChildReference.concatenate( path, new KeyReference(sa.wrappedNodeId) );
						sa.addListener(listener, cis, childPath);
						l.addAccessor(sa);
					}
				} catch (AccessorConstructionException e1) {
					throw new AccessorException( e1 );
				}
			}			
		}

		protected ListenerEntryEx detachListener(Listener listener) throws AccessorException {
			ListenerEntryEx e = listeners;
			ListenerEntryEx p = null;
			while (e!=null) {
				// Found match
				if (e.listener == listener) {
					// The match was the first entry of the linked list
					if (p==null) {
						listeners = e.next;
						return e;
					}
					// Some other entry, unlink e
					p.next = e.next;
					return e;
				}
				p = e;
				e = e.next;
			}
			return null;		
		}
		
		@Override
		public void removeListener(Listener listener) throws AccessorException {
			ListenerEntryEx e = detachListener(listener);
			if (e==null) return;
			MapInterestSet is = (MapInterestSet) e.interestSet;
						
			for (NodeAccessor sa : e.list) {
				InterestSet cis = is.getComponentInterest();
				if (cis!=null) {
					sa.removeListener(listener);				
				}
				cis = is.getComponentInterest( sa.nodeId );
				if (cis!=null) {
					sa.removeListener(listener);				
				}
			}
			e.list.clear();
		}

		@Override
		public int size() throws AccessorException {
			TreeSet<Variant> ids = getIds();
			return ids.size();
		}

		@Override
		public Object getValue(Binding binding) throws AccessorException {
			if (!binding.type().equals(NODES_TYPE)) throw new AccessorException("Wrong binding");
			
			try {
				MapBinding mb = (MapBinding) binding;
				
				TreeSet<Variant> ids = getIds();
				int size = ids.size();
				
				Variant keys[] = ids.toArray( new Variant[size] );
				Object values[] = new NodeRecord[ size ];
				
				for (int i=0; i<size; i++) {
					Variant key = keys[i];
					NodeRecord record = new NodeRecord();
					record.id = key;
					record.labels = new TreeMap<String, String>();
					record.value = new MutableVariant();
					record.children = new ArrayList<Variant>(1);
					
					datasource.getLabels(key, record.labels);
					datasource.getChildren(key, record.children);
					datasource.getValue(key, record.value);
					
					values[i] = record;
				}
				
				return mb.create(keys, values);
			} catch (DatasourceException e) {
				throw new AccessorException(e);
			} catch (BindingException e) {
				throw new AccessorException(e);
			}
		}
		
		@Override
		public void getValue(Binding dstBinding, Object dst) throws AccessorException {
			if (!dstBinding.type().equals(NODES_TYPE)) throw new AccessorException("Wrong binding");
			
			try {
				MapBinding db = (MapBinding) dstBinding;
				
				TreeSet<Variant> ids = getIds();
				int size = ids.size();
				
				Variant keys[] = ids.toArray( new Variant[size] );
				Object values[] = new NodeRecord[ size ];
				
				for (int i=0; i<size; i++) {
					Variant key = keys[i];
					NodeRecord record = new NodeRecord();
					record.id = key;
					record.labels = new TreeMap<String, String>();
					record.value = new MutableVariant();
					record.children = new ArrayList<Variant>(1);
					
					datasource.getLabels(key, record.labels);
					datasource.getChildren(key, record.children);
					datasource.getValue(key, record.value);
					
					values[i] = record;
				}
				
//				return db.create(keys, values);
				throw new AccessorException("not implemented");
			} catch (DatasourceException e) {
				throw new AccessorException(e);
//			} catch (BindingException e) {
//				throw new AccessorException(e);
			}
		}
		
		@Override
		public Object get(Binding keyBinding, Object key, Binding valueBinding)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, IMMUTABLE_VARIANT_BINDING);
				if (!datasource.hasNode(nodeId)) return null;
				RecordBinding rb = (RecordBinding) valueBinding;
				
				TreeMap<String, String> labels = new TreeMap<String, String>();
				MutableVariant value = new MutableVariant();
				ValueContainer container = new ValueContainer();
				ArrayList<Variant> children = new ArrayList<Variant>();

				
				datasource.getLabels(nodeId, labels);
				datasource.getChildren(nodeId, children);
				if (datasource.getValue(nodeId, value)) container.value = value;
				
				Object nodeId2 = Bindings.adapt(nodeId, MUTABLE_VARIANT_BINDING, rb.getComponentBinding(0));
				Object labels2 = Bindings.adapt(labels, LOCALIZEDTEXT_BINDING, rb.getComponentBinding(1)); 
				Object children2 = Bindings.adapt(labels, VARIANT_ARRAY_BINDING, rb.getComponentBinding(2));
				Object value2  = Bindings.adapt(container, OPTIONAL_VARIANT_BINDING, rb.getComponentBinding(3));
				
				return rb.create(nodeId2, labels2, children2, value2);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			} catch (DatasourceException e) {
				throw new AccessorException(e);
			} catch (BindingException e) {
				throw new AccessorException(e);
			}
			
		}

		@Override
		public boolean containsKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				MutableVariant nodeId = (MutableVariant) Bindings.adapt(key, keyBinding, MUTABLE_VARIANT_BINDING);
				return datasource.hasNode(nodeId);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public boolean containsValue(Binding valueBinding, Object value)
				throws AccessorException {
			if (!valueBinding.type().equals( NODE_BINDING.type() )) {
				throw new AccessorException("Unexpected value type");
			}
			
			try {
				RecordAccessor ra = (RecordAccessor) Accessors.getAccessor(valueBinding, value);
				MutableVariant nodeId = (MutableVariant) ra.getFieldValue(0, MUTABLE_VARIANT_BINDING);
				if (!datasource.hasNode(nodeId)) return false;
				
				// Verify fields match
				Object value2 = ra.getValue(valueBinding);
				return valueBinding.equals(value, value2);
			} catch (AccessorConstructionException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public void setValue(Binding mapBinding, Object newMap)
				throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void put(Binding keyBinding, Object key, Binding valueBinding,
				Object value) throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void remove(Binding keyBinding, Object key)
				throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void clear() throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void putAll(Binding keyBinding, Binding valueBinding,
				Map<Object, Object> from) throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void putAll(Binding keyBinding, Binding valueBinding,
				Object[] keys, Object[] values) throws AccessorException {
			throw new WriteAccessException();
		}

		@Override
		public void getAll(Binding keyBinding, Binding valueBinding,
				Map<Object, Object> to) throws AccessorException {
			Set<Variant> ids = getIds();
			int size = ids.size();
			
			try {
				Adapter keyAdapter = Bindings.getAdapter(IMMUTABLE_VARIANT_BINDING, keyBinding);
				Iterator<Variant> iter = ids.iterator(); 
				for (int i=0; i<size; i++) {
					Variant id = iter.next();
					Object value = get(IMMUTABLE_VARIANT_BINDING, id, valueBinding);
					Object key = keyAdapter.adapt( iter.next() );
					to.put(key, value);
				}	
			} catch (AdapterConstructionException e) {
				throw new AccessorException(e);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			}
		}

		@Override
		public void getAll(Binding keyBinding, Binding valueBinding, Object[] keys, Object[] values) throws AccessorException {
			Set<Variant> ids = getIds();
			int size = ids.size();
			if (keys.length<size || values.length<size) throw new AccessorException("array too small");
			
			try {
				Adapter keyAdapter = Bindings.getAdapter(IMMUTABLE_VARIANT_BINDING, keyBinding);
				Iterator<Variant> iter = ids.iterator(); 
				for (int i=0; i<size; i++) {
					Variant id = iter.next();
					values[i] = get(IMMUTABLE_VARIANT_BINDING, id, valueBinding);
					keys[i] = keyAdapter.adapt( iter.next() );
				}	
			} catch (AdapterConstructionException e) {
				throw new AccessorException(e);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			}
		}


		@Override
		public int count(Binding keyBinding, Object from,
				boolean fromInclusive, Object end, boolean endInclusive)
				throws AccessorException {
			throw new AccessorException("Not implemented");
		}

		@Override
		public void getEntries(Binding keyBinding, Object from,
				boolean fromInclusive, Object end, boolean endInclusive,
				ArrayBinding keyArrayBinding, Object dstKeys,
				ArrayBinding valueArrayBinding, Object dstValues)
				throws AccessorException {
			throw new AccessorException("Not implemented");
		}		
		
		@Override
		public Object[] getKeys(Binding keyBinding) throws AccessorException {
			try {
				Set<Variant> ids = getIds();
				int size = ids.size();
				Object[] result = new Object[ size ];
				Adapter keyAdapter = Bindings.getAdapter(IMMUTABLE_VARIANT_BINDING, keyBinding);
				Iterator<Variant> iter = ids.iterator(); 
				for (int i=0; i<size; i++) 
					result[i] = keyAdapter.adapt( iter.next() );
				return result;
			} catch (AdapterConstructionException e) {
				throw new AccessorException(e);
			} catch (AdaptException e) {
				throw new AccessorException(e);
			}
		}

		@Override
		public Object[] getValues(Binding valueBinding)
				throws AccessorException {
			Set<Variant> ids = getIds();
			int size = ids.size();
			Object[] result = new Object[ size ];
			Iterator<Variant> iter = ids.iterator(); 
			for (int i=0; i<size; i++) {
				Variant id = iter.next();
				result[i] = get(IMMUTABLE_VARIANT_BINDING, id, valueBinding);
			}			
			
			return result;
		}

		@Override
		public Object getFirstKey(Binding keyBinding) throws AccessorException {
			try {
				Variant firstId = getIds().first();
				if (firstId==null) return null;			
				return Bindings.adapt(firstId, IMMUTABLE_VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getLastKey(Binding keyBinding) throws AccessorException {
			try {
				Variant lastId = getIds().last();
				if (lastId==null) return null;			
				return Bindings.adapt(lastId, IMMUTABLE_VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getLowerKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (MutableVariant) Bindings.adapt(key, keyBinding, IMMUTABLE_VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant lowerId = ids.lower(nodeId);
				if (lowerId == null) return null;				
				return Bindings.adapt(lowerId, IMMUTABLE_VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getFloorKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (MutableVariant) Bindings.adapt(key, keyBinding, IMMUTABLE_VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant floorId = ids.floor(nodeId);
				if (floorId == null) return null;				
				return Bindings.adapt(floorId, IMMUTABLE_VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getCeilingKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (MutableVariant) Bindings.adapt(key, keyBinding, IMMUTABLE_VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant ceilingId = ids.ceiling(nodeId);
				if (ceilingId == null) return null;				
				return Bindings.adapt(ceilingId, IMMUTABLE_VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public Object getHigherKey(Binding keyBinding, Object key)
				throws AccessorException {
			try {
				Variant nodeId = (Variant) Bindings.adapt(key, keyBinding, IMMUTABLE_VARIANT_BINDING);
				TreeSet<Variant> ids = getIds();
				Variant higherId = ids.higher(nodeId);
				if (higherId == null) return null;				
				return Bindings.adapt(higherId, IMMUTABLE_VARIANT_BINDING, keyBinding);
			} catch (AdaptException e) {
				throw new AccessorException( e );
			}
		}

		@Override
		public void apply(List<Event> cs, LinkedList<Event> rollback) throws AccessorException {
			try {
				boolean makeRollback = rollback != null;
				ArrayList<Event> single = new ArrayList<Event>();
				for (Event e : cs) {
					if (e.reference==null) {
						Event rbe = applyLocal(e, makeRollback);
						if (makeRollback) {
							rbe.reference = e.reference;
							rollback.addFirst( rbe );
						}					
					} else {
						Accessor sa = getComponent(e.reference);
						// Apply changes
						single.clear();
						Event noRefEvent = e.clone(null);
						single.add(noRefEvent);
						sa.apply(single, rollback);
					}
				}
			} catch (AccessorConstructionException ae) {
				throw new AccessorException(ae);
			}				
		}		
		
		Event applyLocal(Event e, boolean makeRollback) throws AccessorException {
			Event rollback = null;
			
			if (e instanceof ValueAssigned) {
				ValueAssigned va = (ValueAssigned) e;
				if (makeRollback) {
					Binding binding = Bindings.getMutableBinding(type());
					rollback = new ValueAssigned(binding, getValue(binding));				
				}
				setValue(va.newValue.getBinding(), va.newValue.getValue());
				return rollback;
			} else if (e instanceof MapEntryAdded) {
				MapEntryAdded ea = (MapEntryAdded) e;
				if (ea.key==null) throw new AccessorException("Cannot apply entry added event because key is missing");
				if (ea.value==null) throw new AccessorException("Cannot apply entry added event because value is missing");
				boolean hadValue = containsKey(ea.key.getBinding(), ea.key.getValue());
				if (hadValue) throw new AccessorException("Could not add entry to key that already existed");
				
				if (makeRollback) {				
					rollback = new MapEntryRemoved( ea.key );
				}
				
				put(ea.key.getBinding(), ea.key.getValue(), ea.value.getBinding(), ea.value.getValue());
				
			} else if (e instanceof MapEntryRemoved) {
				MapEntryRemoved er = (MapEntryRemoved) e;
				
				if (makeRollback) {
					boolean hadValue = containsKey(er.key.getBinding(), er.key.getValue());
					
					if (hadValue) {				
						Object oldValueObj = get(er.key.getBinding(), er.key.getValue(), NODE_BINDING);
						MutableVariant oldKey = er.key;
						MutableVariant oldValue = new MutableVariant(NODE_BINDING, oldValueObj);
						rollback = new MapEntryAdded(oldKey, oldValue);
					} else {
						rollback = new MapEntryRemoved( er.key.clone() );
					}
				}
				
				remove( er.key.getBinding(), er.key.getValue() );
				
			} else throw new AccessorException("Cannot apply "+e.getClass().getName()+" to Map Type");
			
			return rollback;
		}
		
		
	}

	/**
	 * Listener entry that keeps hard references
	 *
	 * @author Toni Kalajainen <toni.kalajainen@vtt.fi>
	 */
	static class ListenerEntryEx {

		public Listener listener;
		public InterestSet interestSet;
		// Path from the listened object 
		public ChildReference path;
		public ListenerEntryEx next;
		ArrayList<NodeAccessor> list = new ArrayList<NodeAccessor>(); 
						
		public ListenerEntryEx(Listener listener, InterestSet interestSet, ChildReference path) {
			if (interestSet==null) throw new IllegalArgumentException();
			this.listener = listener;
			this.interestSet = interestSet;
			this.path = path;
		}
		
		public void addAccessor(NodeAccessor a) {
			list.add(a);
		}
		
		public void removeAccessor(NodeAccessor a) {
			list.remove(a);
		}		
		
		/**
		 * Add new entry to listener list 
		 * 
		 * @param prevValue
		 * @param listener
		 * @param interestSet
		 * @return the new assignment for the list
		 */
		public static ListenerEntryEx link(ListenerEntryEx prevValue, Listener listener, InterestSet interestSet, ChildReference path) {
			ListenerEntryEx e = new ListenerEntryEx(listener, interestSet, path);
			e.next = prevValue;
			return e;
		}
		
		@SuppressWarnings("unchecked")
		public <T extends InterestSet> T getInterestSet() {
			return (T) interestSet;
		}

		
		/**
		 * Remove listener entry from linked list
		 * 
		 * @param firstEntry
		 * @param listener the value to be removed
		 * @return the new assignment for the list
		 */
		public static ListenerEntryEx remove(ListenerEntryEx firstEntry, Listener listener) {
			ListenerEntryEx e = firstEntry;
			ListenerEntryEx p = null;
			while (e!=null) {
				if (e.listener == listener) {
					if (p==null) return e.next;
					p.next = e.next;
					return firstEntry;
				}
				p = e;
				e = e.next;
			}
			return null;
		}	
	}
	
}

